// [[Rcpp::depends(RcppArmadillo)]]
#include <RcppArmadillo.h>

using namespace std;
using namespace Rcpp;
using namespace arma;

/*
// [[Rcpp::export]]
SEXP armadillo_mat_add(List input_params){
  mat A(as<mat>(input_params["A"]));
  mat B(as<mat>(input_params["B"]));
  
  mat C;
  C = A+B;
  return wrap(exp(C));
}
 */


namespace prob_stats{

arma::vec lnormpdf(const arma::vec& x, const arma::vec& mu, const arma::vec& sigma){
  //Constants
  static const double pi = 3.141592653589793238462643383280;
  
  NumericVector sigma_numvec(wrap(sigma));
  NumericVector mu_numvec(wrap(mu));
  NumericVector x_numvec(wrap(x));
  
  vec y;
  
  LogicalVector negative_index;
  negative_index = sigma_numvec <= 0;
  
  sigma_numvec[negative_index] = R_NaN;
  
  vec sigma_vec = as<vec>(sigma_numvec);
  vec mu_vec = as<vec>(mu_numvec);
  vec x_vec = as<vec>(x_numvec);
  
  y = log(1/(sqrt(2*pi)*sigma_vec)) - 0.5*square((x_vec-mu_vec)/sigma_vec);
  
  
  return y;
  
}


arma::vec klgamma(arma::mat pa,arma::mat pb,const double& qa, const double& qb){
  int pa_r = pa.n_rows;
  int pa_n = pa.n_cols;
  int pb_r = pb.n_rows;
  int pb_n = pb.n_cols;
  int n = std::max(pa_n,pb_n);
  int vec_length = pa_r*n;
  
  mat nx = ones<mat>(pa_r,n);
  vec qavec = vectorise(qa*nx);
  vec qbvec = vectorise(qb*nx);
  
  
  if (pa_n==1){
    pa=pa*ones<rowvec>(n);
  }
  if (pb_n==1){
    pb=pb*ones<rowvec>(n);
  }
  
  NumericVector pa_vec(wrap(vectorise(pa)));
  NumericVector pb_vec(wrap(vectorise(pb)));
  NumericVector qa_vec(wrap(qavec));
  NumericVector qb_vec(wrap(qbvec));
  
  NumericVector temp_num(vec_length); 
  
  temp_num = pa_vec*Rcpp::log(pb_vec)-lgamma(pa_vec)
    -qa_vec*log(qb_vec)+lgamma(qa_vec)
    +(pa_vec-qa_vec)*(digamma(pa_vec)-Rcpp::log(pb_vec))
    -(pb_vec-qb_vec)*pa_vec/pb_vec;
    
    vec temp_vec(as<vec>(temp_num));
    mat temp_mat(temp_vec);
    temp_mat.reshape(pa_r,n);
    
    vec row_sum_vec = sum(temp_mat,1);
    return(row_sum_vec);
    
}

}


// [[Rcpp::export]]
SEXP VBDVS_armadillo(List input_params){
  
  // Step 0. Preliminary Inputs
  vec y(as<vec>(input_params["y"]));
  mat x(as<mat>(input_params["x"]));
  int maxiter(input_params["maxiter"]);
  int prior(input_params["prior"]);
  int q(input_params["q"]);
  double delta(input_params["delta"]);
  double c0(input_params["c0"]);
  double h0(input_params["h0"]);
  double g0(input_params["g0"]);
  
  // Step 1. Preliminary Definition
  int Tt = x.n_rows;
  int p = x.n_cols;
  
  // 1.1 Priors
  vec m0 = zeros<vec>(p);
  mat S0 = 4*eye(p,p);
  
  mat ct = ones<mat>(Tt,p);
  mat dt = ones<mat>(Tt,p);
  
  double d0 = 1;
  mat tv_probs = ones<mat>(Tt,p);
  
  mat gt = ones<mat>(Tt,p);
  mat ht = ones<mat>(Tt,p);
  mat tau_1 = zeros<mat>(Tt,p);
  mat tau_0 = zeros<mat>(Tt,p);
  
  double cons;
  vec pi0 = ones<vec>(Tt);
  
  if (prior==1){
    tv_probs.zeros();
  } else {
    tau_1 = gt / ht;
    cons = pow(0.01,2);
    tau_0 = cons*tau_1;
    pi0 = 0.5*pi0;
  }
  
  
  // 1.2 Related auxiliary definitions
  vec at = ones<vec>(Tt);
  double a0=1e-2;
  vec bt = ones<vec>(Tt);
  double b0=1e-2;
  
  // Initialization Vectors
  vec sigma_t = 1e-2*ones<vec>(Tt);
  
  mat mtt = zeros<mat>(p,Tt);
  mat mt1t = zeros<mat>(p,Tt);
  mat mtt1 = zeros<mat>(p,Tt);
  
  cube Stt(p,p,Tt);
  cube Stt1(p,p,Tt);
  cube St1t(p,p,Tt);
  Stt.fill(0);
  Stt1.fill(0);
  St1t.fill(0);
  
  mat lambda_t = zeros<mat>(Tt,p);
  mat q_t = ones<mat>(Tt,p);
  cube Qtilde(p,p,Tt);
  cube Ftilde(p,p,Tt);
  Qtilde.fill(0);
  Ftilde.fill(0);
  
  double Threshold = 1e-4;
  double F_new = 0;
  double F_old = 1;
  int kappa = 1;
  double offset = 0.002;
  
  
  // 1.3 Further auxiliary definitions for later reference
  vec Sx = zeros<vec>(p);
  vec Kt = zeros<vec>(p);
  mat C = zeros<mat>(p,p);
  mat eyeF = zeros<mat>(p,p);
  mat D = zeros<mat>(p,p);
  
  vec l_0 = zeros<vec>(p);
  vec l_1 = zeros<vec>(p);
  
  vec gamma = zeros<vec>(p);
  
  vec s_tinv = zeros<vec>(Tt);
  vec phi = zeros<vec>(Tt);
  
  double temp;
  mat temp_mat = zeros<mat>(1,1);
  
  
  // 2. Implementation of VBDVS
  while (kappa < maxiter && abs(F_new-F_old) > Threshold){
    // 2.1 ======== Update beta using Kalman filter/smoother =========
    
    // Kalman filter loop
    for (int t=0;t<Tt;t++){
      
      if (prior != 1){
        Qtilde.slice(t) = diagmat(1/(q_t.row(t)+lambda_t.row(t)));
        Ftilde.slice(t) = Qtilde.slice(t) * diagmat(q_t.row(t));
      } else {
        Qtilde.slice(t) = diagmat(1/q_t.row(t));
        Ftilde.slice(t) = eye(p,p);
      }
     
      
      if (t==0){
        mtt1.col(t) = Ftilde.slice(t)*m0;
        Stt1.slice(t) = Ftilde.slice(t) * S0 * Ftilde.slice(t).t();
      } else {
        mtt1.col(t) = Ftilde.slice(t)*mtt.col(t-1);
        Stt1.slice(t) = Ftilde.slice(t)*Stt.slice(t-1)*Ftilde.slice(t).t() 
          + Qtilde.slice(t);
      }
      
      Sx = Stt1.slice(t)*x.row(t).t();
      Kt = Sx/(dot(x.row(t),Sx)+sigma_t(t));
      mtt.col(t) = mtt1.col(t) + Kt*(y(t)-dot(x.row(t),mtt1.col(t)));
      Stt.slice(t) = (eye(p,p)-Kt*x.row(t))*Stt1.slice(t);
      
    }
    
    // Fixed Interval Smoother
    
    mt1t.fill(0);
    St1t.fill(0);
    
    mt1t.col(Tt-1) = mtt.col(Tt-1);
    St1t.slice(Tt-1) = Stt.slice(Tt-1);
    
    for (int t=0;t<(Tt-1);t++){
      
      C = Stt.slice(-t+Tt-2)*Ftilde.slice(-t+Tt-2+1)*inv(Stt1.slice(-t+Tt-2+1));
      mt1t.col(-t+Tt-2) = mtt.col(-t+Tt-2) 
        + C*(mt1t.col(-t+Tt-2+1)-mtt1.col(-t+Tt-2+1));
      St1t.slice(-t+Tt-2) = Stt.slice(-t+Tt-2) 
        + C*(St1t.slice(-t+Tt-2+1)-Stt1.slice(-t+Tt-2+1))*C.t();
       
      /*
      C = Stt.slice(t)*Ftilde.slice(t+1)*inv(Stt1.slice(t+1));
      mt1t.col(t) = mtt.col(t) 
        + C*(mt1t.col(t+1)-mtt1.col(t+1));
      St1t.slice(t) = Stt.slice(t) 
        + C*(St1t.slice(t+1)-Stt1.slice(t+1))*C.t();
       */
    }
    
    
    /*
    for (int t=0;t<(Tt-1);t++){
      C = Stt.slice(-t+Tt-2)*Ftilde.slice(-t+Tt-2+1)*inv(Stt1.slice(-t+Tt-2+1));
      mt1t.col(t) = mtt.col(t) 
        + C*(mt1t.col(-t+Tt-2+1)-mtt1.col(-t+Tt-2+1));
      St1t.slice(t) = Stt.slice(t) 
        + C*(St1t.slice(-t+Tt-2+1)-Stt1.slice(-t+Tt-2+1))*C.t();
    }
    */
    
    /*
    for (int t=(Tt-2);t>=0;t--){
      C = Stt.slice(t)*Ftilde.slice(t+1)*inv(Stt1.slice(t+1));
      mt1t.col(t) = mtt.col(t) 
        + C*(mt1t.col(t+1)-mtt1.col(t+1));
      St1t.slice(t) = Stt.slice(t) 
        + C*(St1t.slice(t+1)-Stt1.slice(t+1))*C.t();
    }
    */
    // ============== 2.2 Update Hyperparameters =================
    for (int t=0;t<Tt;t++){
      eyeF = (eye(p,p)-2*Ftilde.slice(t)).t();
      if(t==0){
        D = St1t.slice(t)+mt1t.col(t)*mt1t.col(t).t()+(S0+m0*m0.t())*eyeF;
      } else{
        D = St1t.slice(t)+mt1t.col(t)*mt1t.col(t).t()
        +(St1t.slice(t-1)+mt1t.col(t-1)*mt1t.col(t-1).t())*eyeF;
      }
      
      if (prior==2){
        gt.row(t).fill(g0+0.5);
        ht.row(t) = h0+square(mt1t.col(t)).t()/2;
        tau_1.row(t) = ht.row(t)/gt.row(t);
        tau_0.row(t) = cons*tau_1.row(t);
        
        l_0 = prob_stats::lnormpdf(mt1t.col(t),zeros<vec>(p),sqrt(tau_0.row(t)).t());
        l_1 = prob_stats::lnormpdf(mt1t.col(t),zeros<vec>(p),sqrt(tau_1.row(t)).t());
        gamma = 1/(1+((1-pi0(t))/pi0(t))*exp(l_0-l_1));
        
        pi0(t) = mean(gamma);
        tv_probs.row(t) = gamma.t();
        
        lambda_t.row(t) = (1/(square(1-gamma)%tau_0.row(t).t()+square(gamma)%tau_1.row(t).t())).t();
        
        if (q>0){
          lambda_t(span(t,t),span(0,q)).fill(0);
        }
        
      }
      
      // State variances Q_{t}
      ct.row(t).fill(c0+0.5);
      dt.row(t) = d0+(abs(D.diag())/2).t();
      q_t.row(t) = ct.row(t)/dt.row(t);
      
    }
    
    // =============== 2.3 Update Stochastic Volatilities =============
    s_tinv.fill(0);
    
    for (int t=0;t<Tt;t++){
      temp_mat = x.row(t)*(mt1t.col(t)*mt1t.col(t).t()+St1t.slice(t))*x.row(t).t()
      -2*x.row(t)*mt1t.col(t)*y(t)+(1+offset)*pow(y(t),2);
      //temp_mat = x.row(t)*(x.row(t).t());
      temp = temp_mat(0,0);
      
      if (t==0){
        at(t) = a0+0.5;
        bt(t) = b0+temp/2;
      } else{
        at(t) = delta*at(t-1)+0.5;
        bt(t) = delta*bt(t-1)+temp/2;
      }
      s_tinv(t) = at(t)/bt(t);
    }
    
    // Smooth volatilities
    phi.fill(0);
    phi(Tt-1) = at(Tt-1)/bt(Tt-1);
    
    
    for (int t=0;t<Tt-1;t++){
      phi(-t+Tt-2) = (1-delta)*s_tinv(-t+Tt-2) + delta*phi(-t+Tt-2+1);
      //phi(t) = (1-delta)*s_tinv(t) + delta*phi(t+1);
    }
    
    
    /*
    for (int t=0;t<Tt-1;t++){
      phi(t) = (1-delta)*s_tinv(t) + delta*phi(-t+Tt-2+1);
    }
    */
    
    /*
    for (int t=(Tt-2);t>=0;t--){
      phi(t) = (1-delta)*s_tinv(t) + delta*phi(t+1);
    }
    */
    sigma_t = 1/phi;
    
    // ========= 2.4 Get ELBO and and then go to the next iterations =======
    F_old = F_new;
    F_new = norm(prob_stats::lnormpdf(y,sum(x%mt1t.t(),1),sigma_t),2)
      -sum(prob_stats::klgamma(gt,ht,g0,h0))
      -sum(prob_stats::klgamma(ct,dt,c0,d0));
      
    kappa = kappa+1;
    
  }
  
  Rcpp::List output = Rcpp::List::create(
    Rcpp::Named("beta") = mt1t.t(),
    Rcpp::Named("V") = St1t,
    Rcpp::Named("sigma_t") = sigma_t,
    Rcpp::Named("tv_probs") = tv_probs
  );
  
  return Rcpp::wrap(output);
  
  //return(Rcpp::wrap(Qtilde));
}

